/*
 *	input_raw.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>

#include <sp/spInputPluginP.h>
#include <sp/spPluginMain.h>

#include "plugin_raw.c"

static const char *sp_raw_file_type_list[] = {
    "raw",
    "swap",
    "little",
    "big",
    "ulaw",
    "alaw",
    NULL,
};
static const char *sp_raw_file_desc_list[] = {
    "Raw",
    "Swap Raw",
    "Little Endian",
    "Big Endian",
    "U-law",
    "A-law",
    NULL,
};
static const char *sp_raw_file_filter_list[] = {
    "*",
    "*",
    "*",
    "*",
    "*",
    "*",
    NULL,
};

static spOptions spInitPluginOptionsRaw(void *instance, const char *lang);
static spBool spFreePluginOptionsRaw(void *instance, spOptions options);

static spBool spIsSupportedByPluginRaw(const char *filename);

static spPluginError spOpenPluginRaw(void *instance, const char *filename, const char *mode);
static spBool spClosePluginRaw(void *instance);
static spPluginState spGetPluginStateRaw(void *instance);
static spLong spGetPluginCurrentPositionRaw(void *instance);
static spBool spStopPluginRaw(void *instance);
static spBool spPausePluginRaw(void *instance);
static spBool spRestartPluginRaw(void *instance);

static long spReadPluginRaw(void *instance, char *data, long length);
static spBool spSeekPluginRaw(void *instance, spLong pos);
static spLong spGetPluginTotalLengthRaw(void *instance);

static spInputPluginRec sp_input_plugin_raw = {
    NULL,
    NULL,

    SP_PLUGIN_INPUT,
    "Raw",
    2,
    SP_PLUGIN_PRIORITY_NULL,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginRaw,
    spFreePluginRaw,
    "Raw Input Plugin",
    "Raw Input Plugin Version 0.2\nCopyright(C) Hideki Banno",
    
    spInitPluginInstanceRaw,
    spFreePluginInstanceRaw,
    spInitPluginOptionsRaw,
    spFreePluginOptionsRaw,

    SP_PLUGIN_DEVICE_FILE,
    sp_raw_file_type_list,
    sp_raw_file_desc_list,
    sp_raw_file_filter_list,
    spIsSupportedByPluginRaw,
    spSetPluginFileTypeRaw,  
    spGetPluginFileTypeRaw,
    0,
    spSetPluginSongInfoRaw,
    spGetPluginSongInfoRaw,
    spGetPluginBestSuffixRaw,
    
    NULL,  
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    NULL,
    
    spSetPluginSampleBitRaw,   
    spGetPluginSampleBitRaw,   
    spSetPluginChannelRaw,     
    spGetPluginChannelRaw,     
    spSetPluginSampleRateRaw,  
    spGetPluginSampleRateRaw,  
    spSetPluginOtherInfoRaw,   
    spGetPluginOtherInfoRaw,
    
    spOpenPluginRaw,              
    spClosePluginRaw,             
    spGetPluginStateRaw,          
    spGetPluginCurrentPositionRaw,
    spStopPluginRaw,              
    spPausePluginRaw,             
    spRestartPluginRaw,

    spReadPluginRaw,             
    spSeekPluginRaw,
    spGetPluginTotalLengthRaw,
};

static void updateFileTypeIndex(spRawPluginInstance instance)
{
    int index;
    
    if (!strnone(instance->type)) {
	if ((index = spFindPluginRecFileTypeIndex((spIoPluginRec *)&sp_input_plugin_raw,
						  instance->type)) >= 0) {
	    instance->type_index = index;
	}
    }

    return;
}

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_input_plugin_raw;
}

SP_DECLARE_UNIQUE_GET_PLUGIN_REC(input_raw)

static spOptions spInitPluginOptionsRaw(void *instance, const char *lang)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    spOption options[] = {
	{"-f", "-freq", "sampling frequency for raw data [Hz]",
	     "samp_freq|Sampling Frequency|120|0|Hz|0",
	     SP_TYPE_EDITABLE|SP_TYPE_COMBO|SP_TYPE_DOUBLE, NULL,
	     "8000.0|11025.0|12000.0|16000.0|22050.0|24000.0|32000.0|44100.0|48000.0|96000.0"},
	{"-c", "-channel", "number of channel for raw data",
	     "num_channel|Number of Channel|120",
	     SP_TYPE_EDITABLE|SP_TYPE_COMBO|SP_TYPE_INT, NULL,
	     "1|2|3|4|5|6|8|10|12|16"},
	{"-b", "-bit", "bits per sample for raw data (33: 32bit float)",
	     "bits_per_sample|Bits/Sample|120|0|bits",
	     SP_TYPE_COMBO|SP_TYPE_INT, NULL,
	     "8|@16|24|32|33|64"},
	{"-H", "-head", "header size for raw data [bytes]",
	     "header_size|Header Size|120|0|bytes",
	     SP_TYPE_EDITABLE|SP_TYPE_COMBO|SP_TYPE_LONG, NULL,
	     "0|64|128|256|512|1024"},
	{"-F", "-format", "file format for raw data (raw|swap|little|big|ulaw|alaw)",
	     "file_format|File Format|120",
	     SP_TYPE_COMBO|SP_TYPE_STRING_A, NULL,
	     "raw|swap|little|big|ulaw|alaw"},
    };
    options[0].value = &rinstance->samp_rate;
    options[1].value = &rinstance->num_channel;
    options[2].value = &rinstance->samp_bit;
    options[3].value = &rinstance->head_len;
    options[4].value = rinstance->type;
    

    return spCopyOptions(spArraySize(options), options);
}

static spBool spFreePluginOptionsRaw(void *instance, spOptions options)
{
    spFreeOptions(options);
    return SP_TRUE;
}

static spBool spIsSupportedByPluginRaw(const char *filename)
{
    /* all file is supported, so priority must be low */
    return SP_TRUE;
}

static spPluginError spOpenPluginRaw(void *instance, const char *filename, const char *mode)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    if (strnone(filename)) return SP_PLUGIN_ERROR_OPEN;

    spDebug(10, "spOpenPluginRaw", "filename = %s\n", filename);

    updateFileTypeIndex(rinstance);
    if (rinstance->type_index == SP_RAW_TYPE_ULAW || rinstance->type_index == SP_RAW_TYPE_ALAW) {
	rinstance->samp_bit = 8;
    }

    /* get file length */
    rinstance->file_length = getfilesize(filename, rinstance->head_len)
	/ (long)rinstance->num_channel / (long)(rinstance->samp_bit / 8);
    
    /* open file */
    if (rinstance->file_length <= 0
	|| NULL == (rinstance->fp = spOpenFile(filename, "rb"))) {
	return SP_PLUGIN_ERROR_OPEN;
    }

    /* skip header */
    spSeekFile(rinstance->fp, rinstance->head_len, SEEK_SET);
    rinstance->current_pos = 0;

    spDebug(10, "spOpenPluginRaw", "head_len = %d\n", rinstance->head_len);
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spClosePluginRaw(void *instance)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    if (rinstance->fp != NULL) {
	/* close file */
	spCloseFile(rinstance->fp);
	rinstance->fp = NULL;
	/*rinstance->head_len = 0;*/
	rinstance->current_pos = 0;
    }

    return SP_TRUE;
}

static spPluginState spGetPluginStateRaw(void *instance)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    if (rinstance->current_pos > 0) {
	return SP_PLUGIN_STATE_START;
    }

    return SP_PLUGIN_STATE_STOP;
}

static spLong spGetPluginCurrentPositionRaw(void *instance)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;

    return rinstance->current_pos;
}

static spBool spStopPluginRaw(void *instance)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    if (rinstance->fp != NULL) {
	spSeekFile(rinstance->fp, rinstance->head_len, SEEK_SET);
    }
    rinstance->current_pos = 0;

    return SP_TRUE;
}

static spBool spPausePluginRaw(void *instance)
{

    return SP_TRUE;
}

static spBool spRestartPluginRaw(void *instance)
{

    return SP_TRUE;
}

static long spReadPluginRaw(void *instance, char *data, long length)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    long nread;
    spBool swap_flag = SP_FALSE;

    if (rinstance->fp == NULL) return -1;

    if (rinstance->type_index == SP_RAW_TYPE_SWAP) {
	swap_flag = SP_TRUE;
    } else {
#if BYTE_ORDER == BIG_ENDIAN
	if (rinstance->type_index == SP_RAW_TYPE_LITTLE) {
	    swap_flag = SP_TRUE;
	}
#else
	if (rinstance->type_index == SP_RAW_TYPE_BIG) {
	    swap_flag = SP_TRUE;
	}
#endif
    }
    
    if (rinstance->samp_bit == 8) {
	if (rinstance->type_index == SP_RAW_TYPE_ULAW) {
	    nread = freadulaw((short *)data, length, rinstance->fp);
	} else if (rinstance->type_index == SP_RAW_TYPE_ALAW) {
	    nread = freadalaw((short *)data, length, rinstance->fp);
	} else {
	    nread = freadbyte((short *)data, length, rinstance->fp);
	}
    } else if (rinstance->samp_bit == 64) {
	nread = freaddouble((double *)data, length, swap_flag, rinstance->fp);
    } else if (rinstance->samp_bit == 33) {
	nread = freadfloattod((double *)data, length, swap_flag, rinstance->fp);
    } else if (rinstance->samp_bit == 32) {
	nread = freadlong32((long *)data, length, swap_flag, rinstance->fp);
    } else if (rinstance->samp_bit == 24) {
	nread = freadlong24((long *)data, length, swap_flag, rinstance->fp);
    } else {
	nread = freadshort((short *)data, length, swap_flag, rinstance->fp);
    }
    rinstance->current_pos += (nread / rinstance->num_channel);
    
    return nread;
}

static spBool spSeekPluginRaw(void *instance, spLong pos)
{
    long k, nloop;
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;
    
    if (rinstance->fp == NULL) return SP_FALSE;
    
    spSeekFile(rinstance->fp, rinstance->head_len, SEEK_SET);
    
    /* for a long file */
    nloop = (long)rinstance->num_channel * (long)(rinstance->samp_bit / 8);
    for (k = 0; k < nloop; k++) {
	spSeekFile(rinstance->fp, pos, SEEK_CUR);
    }
    rinstance->current_pos = pos;
    
    return SP_TRUE;
}

static spLong spGetPluginTotalLengthRaw(void *instance)
{
    spRawPluginInstance rinstance = (spRawPluginInstance)instance;

    return rinstance->file_length;
}

