/*
 *	output_flac.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>
#include <sp/spOutputPluginP.h>
#include <sp/spPluginMain.h>

#include <FLAC/stream_encoder.h>
#include <FLAC/metadata.h>

#if !defined(SP_FLAC_OUTPUT_PLUGIN)
#define SP_FLAC_OUTPUT_PLUGIN 1
#endif

#include "plugin_flac.c"

#define SP_FLAC_OUTPUT_BUFFER_LENGTH 1024

static const char *sp_flac_file_type_list[] = {
    "flac",
    "oggflac",
    NULL,
};
static const char *sp_flac_file_desc_list[] = {
    "FLAC",
    "Ogg FLAC",
    NULL,
};
static const char *sp_flac_file_filter_list[] = {
    "*.flac",
    "*.ogg",
    NULL,
};
    
static spOptions spInitPluginOptionsFlac(void *instance, const char *lang);
static spBool spFreePluginOptionsFlac(void *instance, spOptions options);

static spBool spIsSupportedByPluginFlac(const char *filename);

static void *spInitPluginInstanceFlac(const char *lang);
static spBool spFreePluginInstanceFlac(void *instance);

static spPluginError spOpenPluginFlac(void *instance, const char *filename, const char *mode);
static spBool spClosePluginFlac(void *instance);
static spPluginState spGetPluginStateFlac(void *instance);
static spLong spGetPluginCurrentPositionFlac(void *instance);
static spBool spStopPluginFlac(void *instance);
static spBool spPausePluginFlac(void *instance);
static spBool spRestartPluginFlac(void *instance);

static long spWritePluginFlac(void *instance, char *data, long length);
static spBool spFlushPluginFlac(void *instance);


static spOutputPluginRec sp_output_plugin_flac = {
    NULL,
    NULL,

    SP_PLUGIN_OUTPUT,
    "FLAC",
    1,
    SP_PLUGIN_PRIORITY_MIDDLE,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginFlac,
    spFreePluginFlac,
    "Flac Output Plugin",
    "Flac Output Plugin Version 0.1\nCopyright(C) Hideki Banno\nBased on libFLAC (https://xiph.org/flac/api/index.html)",
    
    spInitPluginInstanceFlac,
    spFreePluginInstanceFlac,
    spInitPluginOptionsFlac,
    spFreePluginOptionsFlac,

    SP_PLUGIN_DEVICE_FILE,
    sp_flac_file_type_list,
    sp_flac_file_desc_list,
    sp_flac_file_filter_list,
    spIsSupportedByPluginFlac,
    spSetPluginFileTypeFlac,  
    spGetPluginFileTypeFlac,
    (SP_SONG_TITLE_MASK | SP_SONG_ARTIST_MASK | SP_SONG_COMMENT_MASK
     | SP_SONG_GENRE_MASK | SP_SONG_RELEASE_MASK | SP_SONG_COPYRIGHT_MASK
     | SP_SONG_TRACK_MASK | SP_SONG_ISRC_MASK | SP_SONG_ALBUM_MASK | SP_SONG_SOFTWARE_MASK),
    spSetPluginSongInfoFlac,
    spGetPluginSongInfoFlac,
    spGetPluginBestSuffixFlac,
    
    NULL,  
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    NULL,
	    
    spSetPluginSampleBitFlac,   
    spGetPluginSampleBitFlac,   
    spSetPluginChannelFlac,     
    spGetPluginChannelFlac,     
    spSetPluginSampleRateFlac,  
    spGetPluginSampleRateFlac,  
    spSetPluginOtherInfoFlac,   
    spGetPluginOtherInfoFlac,
    
    spOpenPluginFlac,              
    spClosePluginFlac,             
    spGetPluginStateFlac,          
    spGetPluginCurrentPositionFlac,
    spStopPluginFlac,              
    spPausePluginFlac,             
    spRestartPluginFlac,

    spWritePluginFlac,             
    spFlushPluginFlac,
    NULL,
};

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_output_plugin_flac;
}

SP_DECLARE_UNIQUE_GET_PLUGIN_REC(output_flac)

static void *spInitPluginInstanceFlac(const char *lang)
{
    FLAC__StreamEncoder *encoder;
    spFlacPluginInstance instance;

    if ((encoder = FLAC__stream_encoder_new()) == NULL) {
	spDebug(1, "spInitPluginInstanceFlac", "ERROR: allocating encoder\n");
	return NULL;
    }
    
    instance = xalloc(1, struct _spFlacPluginInstance);
    instance->encoder = encoder;
    instance->encode_comp_level = 5;
    instance->encode_bits_per_sample = 0;
    instance->encode_mult_weight = 1;
    instance->encode_div_weight = 1;
    
    instance->metadata = NULL;
    
    instance->buffer_copy_pos = 0;
    instance->output_buffer_length = 0;
    instance->output_buffer = NULL;
    
    instance->total_samples = 0;
    instance->sample_rate = 0;
    instance->channels = 0;
    instance->bits_per_sample = 0;
    instance->max_blocksize = 0;

    instance->is_ogg = SP_FALSE;
    spInitSongInfoV2(&instance->song_info);
    
    instance->current_pos = 0;

    spDebug(10, "spInitPluginInstanceFlac", "done\n");
    
    return (void *)instance;
}

static spBool spFreePluginInstanceFlac(void *instance)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;

    if (pinstance == NULL) return SP_FALSE;

    if (pinstance->encoder != NULL) {
	FLAC__stream_encoder_delete(pinstance->encoder);
    }
    
    xfree(pinstance);
    
    return SP_TRUE;
}

static spOptions spInitPluginOptionsFlac(void *instance, const char *lang)
{
    int i;
    int num_option;
    spOption *options;
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
#if defined(SP_SUPPORT_UTF8_STRING)
#include "options_UTF8JP.h"
    const char *ja_lang = "ja_JP.utf8";
#elif defined(_WIN32) || defined(MACOS)
#include "options_SJIS.h"
    const char *ja_lang = "ja_JP.SJIS";
#else
#include "options_EUCJP.h"
    const char *ja_lang = "ja_JP.eucJP";
#endif
    spOption options_en[] = {
	{"-level", NULL, "compression level (0: fast, 8: best but slow)", "comp_level|Compression Level|130",
	     SP_TYPE_COMBO|SP_TYPE_LONG|SP_TYPE_SHOW_TIP, NULL, "0|1|2|3|4|@5|6|7|8"},
	{"-b", "-bit", "bits per sample", "bits_per_sample|Bits/Sample|130",
	     SP_TYPE_COMBO|SP_TYPE_INT|SP_TYPE_SHOW_TIP, NULL, "@16|24"},
    };

    if (spEqLanguage(lang, ja_lang)) {
	num_option = spArraySize(options_ja);
	options = options_ja;
    } else {
	num_option = spArraySize(options_en);
	options = options_en;
    }
    i = 0;
    options[i++].value = &pinstance->encode_comp_level;
    
    spDebug(10, "spInitPluginOptionsFlac", "current bits_per_sample = %d\n", pinstance->bits_per_sample);
    pinstance->encode_bits_per_sample = pinstance->bits_per_sample;
    options[i++].value = &pinstance->encode_bits_per_sample;

    return spCopyOptions(num_option, options);
}

static spBool spFreePluginOptionsFlac(void *instance, spOptions options)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    spFreeOptions(options);
    spDebug(10, "spFreePluginOptionsFlac", "encode_bits_per_sample = %d\n", pinstance->encode_bits_per_sample);
    return SP_TRUE;
}

static spBool spIsSupportedByPluginFlac(const char *filename)
{
    const char *p;
    
    if (filename != NULL && (p = strrchr(filename, '.')) != NULL) {
	if (strcaseeq(p, ".flac") || strcaseeq(p, ".ogg") || strcaseeq(p, ".oga")) {
	    return SP_TRUE;
	}
    }

    return SP_FALSE;
}

static spBool setMetadataVorbisComment(FLAC__StreamMetadata *metadata, spSongInfoV2 *song_info)
{
    FLAC__StreamMetadata_VorbisComment_Entry entry;
    spConverter conv;
    char buf[SP_SONG_INFO_SIZE];
    
    if (song_info->info_mask & SP_SONG_TRACK_MASK) {
	sprintf(buf, "%d", (int)song_info->track);
	FLAC__metadata_object_vorbiscomment_entry_from_name_value_pair(&entry, "TRACKNUMBER", buf);
	FLAC__metadata_object_vorbiscomment_append_comment(metadata, entry, /*copy=*/false);
    }

    if ((conv = spOpenConverter(NULL, "UTF-8")) != NULL) {
	int i;
	char *name;
	char *iptr, *ostr;
	unsigned long mask_done = 0L;
	
	for (i = 0; sp_flac_song_info_table[i].mask; i++) {
	    iptr = (char *)song_info + sp_flac_song_info_table[i].offset;
	    
	    if (!(mask_done & sp_flac_song_info_table[i].mask)
		&& (song_info->info_mask & sp_flac_song_info_table[i].mask) && !strnone(iptr)) {
		mask_done |= sp_flac_song_info_table[i].mask;
		name = sp_flac_song_info_table[i].name;
		spDebug(10, "setMetadataVorbisComment", "add tag: i = %d, name = %s, iptr = %s\n", i, name, iptr);
	    
		if ((ostr = xspConvert(conv, iptr)) != NULL) {
		    FLAC__metadata_object_vorbiscomment_entry_from_name_value_pair(&entry, name, ostr);
		    FLAC__metadata_object_vorbiscomment_append_comment(metadata, entry, /*copy=*/false);
		    xfree(ostr);
		}
	    }
	}

	spCloseConverter(conv);
	spDebug(10, "setMetadataVorbisComment", "set string comment done\n");
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool writeCurrentBufferFlac(spFlacPluginInstance pinstance)
{
    FLAC__bool ok;
    unsigned l;
    
    if (pinstance->encode_mult_weight != 1 || pinstance->encode_div_weight != 1) {
	for (l = 0; l < pinstance->buffer_copy_pos; l++) {
	    if (pinstance->encode_mult_weight != 1) {
		pinstance->output_buffer[l] *= pinstance->encode_mult_weight;
	    } else {
		pinstance->output_buffer[l] = (FLAC__int32)spRound((double)pinstance->output_buffer[l] / (double)pinstance->encode_div_weight);
	    }
	}
    }
    
    ok = FLAC__stream_encoder_process_interleaved(pinstance->encoder, pinstance->output_buffer, pinstance->buffer_copy_pos);
    pinstance->buffer_copy_pos = 0;

    if (ok) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static void progressCallback(const FLAC__StreamEncoder *encoder, FLAC__uint64 bytes_written, FLAC__uint64 samples_written,
			     unsigned frames_written, unsigned total_frames_estimate, void *client_data)
{
    spDebug(100, "progressCallback", "wrote %lu bytes, %lu samples, %u/%u frames\n",
	    (unsigned long)bytes_written, (unsigned long)samples_written, frames_written, total_frames_estimate);
}

static spPluginError spOpenPluginFlac(void *instance, const char *filename, const char *mode)
{
    FLAC__StreamEncoderInitStatus init_status;
    FLAC__bool ok = true;
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    /* open file */
    if (spCheckSuffixFlac(filename, &pinstance->is_ogg) == SP_FALSE) {
	return SP_PLUGIN_ERROR_OPEN;
    }
    spDebug(10, "spOpenPluginFlac", "open file done: %s\n", filename);

    ok &= FLAC__stream_encoder_set_verify(pinstance->encoder, true);
    ok &= FLAC__stream_encoder_set_compression_level(pinstance->encoder, pinstance->encode_comp_level);
    ok &= FLAC__stream_encoder_set_channels(pinstance->encoder, pinstance->channels);
    ok &= FLAC__stream_encoder_set_bits_per_sample(pinstance->encoder, pinstance->encode_bits_per_sample);
    ok &= FLAC__stream_encoder_set_sample_rate(pinstance->encoder, pinstance->sample_rate);
    ok &= FLAC__stream_encoder_set_total_samples_estimate(pinstance->encoder, pinstance->total_samples);

    if (!ok) {
	return SP_PLUGIN_ERROR_OPEN;
    }

    if ((pinstance->metadata = FLAC__metadata_object_new(FLAC__METADATA_TYPE_VORBIS_COMMENT)) != NULL) {
	if (setMetadataVorbisComment(pinstance->metadata, &pinstance->song_info) == SP_TRUE) {
	    if (FLAC__stream_encoder_set_metadata(pinstance->encoder, &pinstance->metadata, 1)) {
		spDebug(10, "spOpenPluginFlac", "FLAC__stream_encoder_set_metadata OK\n");
	    } else {
		spDebug(10, "spOpenPluginFlac", "FLAC__stream_encoder_set_metadata failed\n");
	    }
	}
    }

    if (pinstance->is_ogg) {
        init_status = FLAC__stream_encoder_init_ogg_file(pinstance->encoder, filename, progressCallback, pinstance);
    } else {
        init_status = FLAC__stream_encoder_init_file(pinstance->encoder, filename, progressCallback, pinstance);
    }
    
    if (init_status != FLAC__STREAM_ENCODER_INIT_STATUS_OK) {
	spDebug(1, "spOpenPluginFlac", "ERROR: initializing encoder: %s\n", FLAC__StreamEncoderInitStatusString[init_status]);
	return SP_PLUGIN_ERROR_OPEN;
    }

    if (pinstance->encode_bits_per_sample == pinstance->bits_per_sample) {
	pinstance->encode_mult_weight = 1;
	pinstance->encode_div_weight = 1;
    } else {
	if (pinstance->bits_per_sample > pinstance->encode_bits_per_sample) {
	    pinstance->encode_div_weight = POW2(pinstance->bits_per_sample - pinstance->encode_bits_per_sample);
	} else {
	    pinstance->encode_mult_weight = POW2(pinstance->encode_bits_per_sample - pinstance->bits_per_sample);
	}
    }
    spDebug(10, "spOpenPluginFlac", "mult_weight = %ld, div_weight = %ld\n",
	    pinstance->encode_mult_weight, pinstance->encode_div_weight);

    pinstance->buffer_copy_pos = 0;
    pinstance->output_buffer_length = SP_FLAC_OUTPUT_BUFFER_LENGTH;
    pinstance->output_buffer = xalloc(pinstance->output_buffer_length * pinstance->channels, FLAC__int32);
     
    pinstance->current_pos = 0;
    
    spDebug(10, "spOpenPluginFlac", "done\n");
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spClosePluginFlac(void *instance)
{
    FLAC__bool ok;
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;

    if (pinstance->output_buffer != NULL && pinstance->buffer_copy_pos > 0) {
	writeCurrentBufferFlac(pinstance);
    }

    ok = FLAC__stream_encoder_finish(pinstance->encoder);

    if (pinstance->metadata != NULL) {
	FLAC__metadata_object_delete(pinstance->metadata);
	pinstance->metadata = NULL;
    }
    
    if (pinstance->output_buffer != NULL) {
	xfree(pinstance->output_buffer);
	pinstance->output_buffer = NULL;
    }
    
    pinstance->total_samples = pinstance->current_pos;
    pinstance->current_pos = 0;

    if (ok) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spPluginState spGetPluginStateFlac(void *instance)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    if (pinstance->current_pos > 0) {
	return SP_PLUGIN_STATE_START;
    }

    return SP_PLUGIN_STATE_STOP;
}

static spLong spGetPluginCurrentPositionFlac(void *instance)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    return (spLong)pinstance->current_pos;
}

static spBool spStopPluginFlac(void *instance)
{
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;
    
    pinstance->current_pos = 0;
    pinstance->buffer_copy_pos = 0;

    return SP_TRUE;
}

static spBool spPausePluginFlac(void *instance)
{
    return SP_TRUE;
}

static spBool spRestartPluginFlac(void *instance)
{
    return SP_TRUE;
}

static long spWritePluginFlac(void *instance, char *data, long length)
{
    unsigned l;
    long offset;
    long length2;
    long nremain;
    long copy_length;
    short *sdata;
    long *ldata;
    spFlacPluginInstance pinstance = (spFlacPluginInstance)instance;    
    
    spDebug(80, "spWritePluginFlac", "in: length = %ld\n", length);

    sdata = (short *)data;
    ldata = (long *)data;

    offset = 0;
    length2 = length / pinstance->channels;
    nremain = length2;
    
    while (nremain > 0) {
	if (pinstance->buffer_copy_pos >= pinstance->output_buffer_length) {
	    writeCurrentBufferFlac(pinstance);
	}

	copy_length = MIN(nremain, (long)(pinstance->output_buffer_length - pinstance->buffer_copy_pos));
	
	if (sizeof(long) == sizeof(FLAC__int32)
	    && pinstance->encode_bits_per_sample >= 24 && pinstance->bits_per_sample >= 24) {
	    memcpy(pinstance->output_buffer + pinstance->buffer_copy_pos * pinstance->channels,
		   ldata + offset * pinstance->channels, copy_length * pinstance->channels * sizeof(long));
	} else {
	    for (l = 0; l < copy_length * pinstance->channels; l++) {
		if (pinstance->bits_per_sample >= 24) {
		    pinstance->output_buffer[pinstance->buffer_copy_pos * pinstance->channels + l]
			= (FLAC__int32)ldata[offset * pinstance->channels + l];
		} else {
		    pinstance->output_buffer[pinstance->buffer_copy_pos * pinstance->channels + l]
			= (FLAC__int32)sdata[offset * pinstance->channels + l];
		}
	    }
	}
	pinstance->buffer_copy_pos += copy_length;

	nremain -= copy_length;
	offset += copy_length;
    }
    
    pinstance->current_pos += offset;
    spDebug(80, "spWritePluginFlac", "done: current_pos = %ld\n", pinstance->current_pos);

    return offset * pinstance->channels;
}

static spBool spFlushPluginFlac(void *instance)
{
    return SP_TRUE;
}

