#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spMain.h>
#include <sp/spBaseLib.h>

#include <sp/spPlugin.h>
#include <sp/spOutputPlugin.h>
#include <sp/spInputPlugin.h>

#define SPPLAY_HELP_MESSAGE "play sound file"

#define SPPLAY_MIN_BUF_LENGTH 128

static spBool help_flag;
static spBool quiet_flag;
static spBool second_flag;
static spBool supported_only;
static spBool display_song_info;
static int debug_level = -1;
static char i_param_file[SP_MAX_PATHNAME] = "";
static char o_param_file[SP_MAX_PATHNAME] = "";

/*static char *plugin_search_path = NULL;*/
static char *plugin_search_path;
static char o_plugin_name[SP_MAX_LINE] = "";

static double samp_freq = 0.0;
static int num_channel = 0;
static int samp_bit = 0;

static long buffer_length = 0;
static double offset_s = 0.0;
static double length_s = 0.0;
static double output_gain = 1.0;
static double interval = 0.0;
static char output_file[SP_MAX_PATHNAME] = "";
static char output_format[SP_MAX_PATHNAME] = "";

static spOptions options;
static spOption option[] = {
    {"-o", "-offset", "offset time [s]", "offset",
	 SP_TYPE_DOUBLE, &offset_s, NULL},
    {"-l", "-length", "length of playing [s]", "length",
	 SP_TYPE_DOUBLE, &length_s, NULL},
    {"-O", "-output", "output file (default format: wav)", NULL,
	 SP_TYPE_STRING_A, output_file, NULL},
    {"-g", "-gain", "output gain", "output_gain",
	 SP_TYPE_DOUBLE, &output_gain, "1.0"},
    {"-i", "-interval", "interval time between sound files for file output", "interval",
	 SP_TYPE_DOUBLE, &interval, "0.0"},
    {"-buf", NULL, "buffer length for audio output", "buffer_length",
	 SP_TYPE_LONG, &buffer_length, "1024"},
    {"-op", "-oplugin", "plugin name for output", "output_plugin",
	 SP_TYPE_STRING_A, o_plugin_name, "output_audio"},
    {"-OF", NULL, "output file format (`help' lists supported formats)", NULL,
	 SP_TYPE_STRING_A, output_format, NULL},
    {"-path", NULL, "extra plugin search path", "plugin_path",
	 SP_TYPE_STRING, &plugin_search_path, NULL},
    {"-song", NULL, "display song info", "display_song_info",
	 SP_TYPE_BOOLEAN, &display_song_info, SP_FALSE_STRING},
    {"-so", NULL, "play supported file only", "supported_only",
	 SP_TYPE_BOOLEAN, &supported_only, SP_FALSE_STRING},
    {"-s", "-sec", "display time in second", "time_second",
	 SP_TYPE_BOOLEAN, &second_flag, SP_FALSE_STRING},
    {"-q", "-quiet", "quiet mode", NULL,
	 SP_TYPE_BOOLEAN, &quiet_flag, SP_FALSE_STRING},
    {"-iparam", NULL, "input parameter file", NULL,
	 SP_TYPE_STRING_A, i_param_file, NULL},
    {"-oparam", NULL, "output parameter file", NULL,
	 SP_TYPE_STRING_A, o_param_file, NULL},
    {"-debug", NULL, "debug level", NULL,
	 SP_TYPE_INT, &debug_level, NULL},
    {"-h", "-help", "display this message", NULL,
	 SP_TYPE_BOOLEAN, &help_flag, SP_FALSE_STRING},
    /* options for input_raw */
    SpNullOption("-f", "-freq", SP_TYPE_NONE),
    SpNullOption("-c", "-channel", SP_TYPE_NONE),
    SpNullOption("-b", "-bit", SP_TYPE_NONE),
    SpNullOption("-H", "-head", SP_TYPE_NONE),
    SpNullOption("-F", "-format", SP_TYPE_NONE),
};

static const char *filelabel[] = {
    "<input file...>",
};

static void usage(spPlugin *i_plugin, spPlugin *o_plugin)
{
    spPrintHelpHeader(SPPLAY_HELP_MESSAGE);
    spPrintUsageHeader();
    
    if (i_plugin != NULL) {
	spPrintPluginOptions(i_plugin);
    } else {
	spPluginUsage("input_raw");
    }
    
    if (o_plugin != NULL) {
	spPrintPluginOptions(o_plugin);
    }
    
    if (i_plugin != NULL) {
	spCloseFilePlugin(i_plugin);
    }
    if (o_plugin != NULL) {
	spCloseFilePlugin(o_plugin);
    }
    spPrintUsage();
    
    return;
}

int spMain(int argc, char *argv[])
{
    int i;
    int samp_byte;
    long k;
    long len;
    long nread;
    long nwrite;
    long current_buffer_length;
    spLong offset;
    spLong length;
    spLong total_length;
    spLong current_length;
    spLong interval_length;
    spLong position;
    double *buf;
    const char *i_filename;
    char *o_filename;
    spPluginDeviceType device_type;
    spWaveInfo wave_info;
    spSongInfo song_info;
    spPluginError plugerror;
    spPlugin *i_plugin = NULL;
    spPlugin *o_plugin = NULL;
    char orig_o_plugin_name[SP_MAX_LINE];

    spSetSetup("~/.spplay");

    options = spGetOptions(argc, argv, option, filelabel);
    spIgnoreUnknownOption(options, SP_TRUE);
    spGetOptionsValue(argc, argv, options);
    spSetDebugLevel(debug_level);

    /* read parameter file */
    if (!strnone(i_param_file)) {
	spMessage("Input Parameter File: %s\n", i_param_file);
	spReadSetup(i_param_file, options);
    }
    
    if (!strnone(plugin_search_path)) {
	spDebug(10, "spMain", "plugin_search_path: %s\n", plugin_search_path);
	spAppendPluginSearchPath(plugin_search_path);
    }
    
    buf = xalloc(buffer_length, double);

    if ((i_filename = spGetFile(options)) == NULL) {
	if (help_flag == SP_TRUE) {
	    usage(NULL, NULL);
	} else {
	    spError(1, "Not enough files.\n");
	}
    }

    device_type = SP_PLUGIN_DEVICE_AUDIO;
    strcpy(orig_o_plugin_name, o_plugin_name);
    
    do {
	/* initialize option */
	spInitWaveInfo(&wave_info);
	wave_info.samp_bit = samp_bit;
	wave_info.num_channel = num_channel;
	wave_info.samp_rate = samp_freq;
	spInitSongInfo(&song_info);
	
	/* open input file */
	plugerror = SP_PLUGIN_ERROR_SUCCESS;
	if ((i_plugin = spOpenFilePluginArg(NULL, i_filename, "r",
					    SP_PLUGIN_DEVICE_FILE,
					    &wave_info, &song_info, argc, argv, &plugerror)) == NULL) {
	    if (supported_only == SP_TRUE) {
		if (quiet_flag == SP_FALSE) {
		    spWarning("This file is not supported: %s\n", i_filename);
		}
		continue;
	    }
	    
	    spDebug(10, "spMain", "can't find suitable plugin\n");
	    plugerror = SP_PLUGIN_ERROR_SUCCESS;
	    if ((i_plugin = spOpenFilePluginArg("input_raw", i_filename, "r",
						SP_PLUGIN_DEVICE_FILE,
						&wave_info, NULL, argc, argv, &plugerror)) == NULL) {
		/*spError(1, "Can't find plugins.\n");*/
		continue;
	    }
	}
	total_length = spGetPluginTotalLength(i_plugin);
	samp_byte = wave_info.samp_bit / 8;

	/* update information for next loop */
	samp_bit = wave_info.samp_bit;
	num_channel = wave_info.num_channel;
	samp_freq = wave_info.samp_rate;
        spDebug(10, "spMain", "input: samp_bit = %d, num_channel = %d, samp_freq = %f, output_gain = %f\n",
                samp_bit, num_channel, samp_freq, output_gain);

        current_buffer_length = (buffer_length / (long)num_channel) * num_channel;
	
	/* open output plugin */
	if (!strnone(output_file)) {
	    if (o_plugin == NULL) {
		device_type = SP_PLUGIN_DEVICE_FILE;
		o_filename = output_file;
		if (streq(o_plugin_name, "output_audio")) {
                    if (!streq(orig_o_plugin_name, "output_audio")) {
                        strcpy(o_plugin_name, orig_o_plugin_name);
                    } else {
                        strcpy(o_plugin_name, "output_wav");
                    }
		}
	    }
	} else {
	    strcpy(o_plugin_name, orig_o_plugin_name);
	    
	    if (o_plugin != NULL) {
		spCloseFilePlugin(o_plugin);
		o_plugin = NULL;
	    }
	    device_type = SP_PLUGIN_DEVICE_AUDIO;

	    /* buffer length limitation to prevent long waiting on DOS */
	    len = (long)(total_length / 32) * wave_info.num_channel;
	    if (len <= current_buffer_length) {
		wave_info.buffer_size =
		    MAX(len / SPPLAY_MIN_BUF_LENGTH, 1)
			* SPPLAY_MIN_BUF_LENGTH * samp_byte;
	    } else {
		wave_info.buffer_size = current_buffer_length * samp_byte;
	    }
	    o_filename = NULL;
	}
	if (!strnone(output_format)) {
	    strcpy(wave_info.file_type, output_format);
	} else {
	    strcpy(wave_info.file_type, "");
	}
	strcpy(wave_info.file_desc, "");
	strcpy(wave_info.file_filter, "");
	if (o_plugin == NULL) {
	    if ((o_plugin = spOpenFilePluginArg(o_plugin_name, o_filename, "w",
						device_type,
						&wave_info, &song_info, argc, argv, &plugerror)) == NULL) {
		spDebug(1, NULL, "plugerror = %d\n", plugerror);
		if (plugerror == SP_PLUGIN_ERROR_FILE_TYPE) {
		    fprintf(stderr, "The file type is not supported.\n");
		} else if (plugerror == SP_PLUGIN_ERROR_SAMP_BIT) {
		    fprintf(stderr, "The bits/sample is not supported.\n");
		} else if (plugerror == SP_PLUGIN_ERROR_NUM_CHANNEL) {
		    fprintf(stderr, "The number of channels is not supported.\n");
		} else if (plugerror == SP_PLUGIN_ERROR_SAMP_RATE) {
		    fprintf(stderr, "The sampling rate is not supported.\n");
		} else {
		    fprintf(stderr, "Can't open output plugin.\n");
		}
		spExit(1);
	    }
            spDebug(10, "spMain", "output: samp_bit = %d, num_channel = %d, samp_freq = %f\n",
                    wave_info.samp_bit, wave_info.num_channel, wave_info.samp_rate);
	
	    spDebug(10, NULL, "output plugin: %s\n", spGetPluginName(o_plugin));
	    spDebug(10, NULL, "output buffer size: %d\n", wave_info.buffer_size);

	    if (help_flag == SP_TRUE) {
		usage(i_plugin, o_plugin);
	    }
	} else {
	    if (samp_bit != wave_info.samp_bit) {
		spWarning("Bits/sample (%d) is different from target file's one (%d)\n", wave_info.samp_bit, samp_bit);
	    }
	    if (num_channel != wave_info.num_channel) {
		spWarning("Number of channels (%d) is different from target file's one (%d)\n",
			  wave_info.num_channel, num_channel);
	    }
	    if (samp_freq != wave_info.samp_rate) {
		spWarning("Sampling rate (%f) is different from target file's one (%f)\n",
			  wave_info.samp_rate, samp_freq);
	    }
	}
	
	if (quiet_flag == SP_FALSE) {
	    char *string;

	    printf("Input Filename: %s\n", i_filename);
	    
	    if ((string = xspGetPluginFileType(i_plugin, SP_TRUE)) != NULL) {
		printf("File Type: %s\n", string);
		xfree(string);
	    }
	    printf("Sampling Rate: %f\n", samp_freq);
	    printf("Number of Channels: %d\n", num_channel);
	    printf("Bits/Sample: %d\n", samp_bit);
	    printf("Total Length: %ld\n", (long)total_length);

	    if (display_song_info == SP_TRUE) {
		spShowSongInfo(&song_info, NULL, stdout);
	    }
	}
	
	spDebug(10, NULL, "input plugin: %s\n", spGetPluginName(i_plugin));
	spDebug(10, NULL, "total length: %ld\n", total_length);

	offset = (spLong)spRound(offset_s * wave_info.samp_rate);
	length = (spLong)spRound(length_s * wave_info.samp_rate) * wave_info.num_channel;
	interval_length = (spLong)spRound(interval * wave_info.samp_rate) * wave_info.num_channel;

	if (offset > 0) {
	    spSeekPlugin(i_plugin, offset);
	}

	spDebug(10, NULL, "main loop\n");
	
	current_length = 0;
	for (i = 0;; i++) {
	    if (quiet_flag == SP_FALSE) {
		if ((position = spGetPluginCurrentPosition(o_plugin)) < 0) {
		    position = current_length / wave_info.num_channel;
		}
		spDebug(10, NULL, "position = %ld\n", position);
		
		if (second_flag == SP_TRUE) {
		    spMessage("Time: %5.2f [s] %c%c",
			      (double)(position + offset) / wave_info.samp_rate, 
			      (char)13, (char)0);
		} else {
		    int hour, minute, second;
		    
		    second = (int)((double)(position + offset) / wave_info.samp_rate);
		    minute = second / 60;
		    hour = minute / 60;
		    spDebug(10, NULL, "%.2d:%.2d:%.2d\n", hour, minute, second);
		    minute %= 60;
		    second %= 60;
		    
		    spMessage("Time: %.2d:%.2d:%.2d %c%c",
			      hour, minute, second, (char)13, (char)0);
		}
	    }

	    if ((nread = spReadPluginDouble(i_plugin, buf, current_buffer_length)) <= 0) {
		break;
	    }
	    current_length += nread;
	    spDebug(10, NULL, "nread = %ld, current_length = %ld, length = %ld\n",
		    nread, current_length, length);
	    
	    if (length > 0 && current_length >= length) {
		nwrite = spWritePluginDoubleWeighted(o_plugin, buf,
                                                     nread - (long)(current_length - length), output_gain);
		break;
	    } else {
		nwrite = spWritePluginDoubleWeighted(o_plugin, buf, nread, output_gain);
	    }
#if 0
            for (k = 0; k < nwrite; k++) {
                printf("%f\n", buf[k]);
            }
#endif
	    spDebug(10, NULL, "nwrite = %ld\n", nwrite);
	}

	if (interval_length > 0) {
            for (k = 0; k < current_buffer_length; k++) {
                buf[k] = 0.0;
            }
	    
	    current_length = 0;
	    while (1) {
		nread = current_buffer_length;
		current_length += nread;
		if (interval_length > 0 && current_length >= interval_length) {
		    nwrite = spWritePluginDoubleWeighted(o_plugin, buf,
                                                         nread - (long)(current_length - interval_length), 1.0);
		    break;
		} else {
		    if ((nwrite = spWritePluginDoubleWeighted(o_plugin, buf, nread, 1.0)) <= 0) {
			break;
		    }
		}
	    }
	}

	spDebug(10, NULL, "main loop done\n");
	
	/* close plugin */
	spCloseFilePlugin(i_plugin);

	if (device_type == SP_PLUGIN_DEVICE_FILE) {
	    printf("\nOutput Filename: %s\n", o_filename);
	    printf("File Type: %s\n", wave_info.file_type);
	    printf("Sampling Rate: %f\n", wave_info.samp_rate);
	    printf("Number of Channels: %d\n", wave_info.num_channel);
	    printf("Bits/Sample: %d\n", wave_info.samp_bit);
	    /*break;*/
	}
	
    } while ((i_filename = spGetFile(options)) != NULL);

    if (o_plugin != NULL) {
	spCloseFilePlugin(o_plugin);
    }
    
    xfree(buf);

    /* write parameter file */
    if (!strnone(o_param_file)) {
	spMessage("Output Parameter File: %s\n", o_param_file);
	spWriteSetup(o_param_file, options);
    } else {
	spWriteGlobalSetup();
    }
    
    return 0;
}
